/*
 * OptimPack3.i --
 *
 *	Yorick interface to OptimPack library.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 2003-2007, Eric Thibaut.
 *
 *	This file is part of OptimPack.
 *
 *	OptimPack is  free software; you can redistribute  it and/or modify
 *	it under the  terms of the GNU General  Public License as published
 *	by the Free  Software Foundation; either version 2  of the License,
 *	or (at your option) any later version.
 *
 *	OptimPack is  distributed in the hope  that it will  be useful, but
 *	WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 *	MERCHANTABILITY or  FITNESS FOR A PARTICULAR PURPOSE.   See the GNU
 *	General Public License for more details.
 *
 *	You should have  received a copy of the  GNU General Public License
 *	along  with   OptimPack;  if  not,  write  to   the  Free  Software
 *	Foundation,  Inc.,   51  Franklin  St,  Fifth   Floor,  Boston,  MA
 *	02110-1301 USA
 *
 *-----------------------------------------------------------------------------
 *
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

/*---------------------------------------------------------------------------*/
/* INITIALISATION OF OPTIMPACK */

/* First of all OPTIMPACK requires hash-tables (and some other functions)
   provided by Yeti: */
if (is_func(current_include) != 2) {
  include, "yeti.i", 3;
  if (is_func(h_new) != 2) {
    error, "Yeti is mandatory to run OPTIMPACK.";
  }
}

local OPTIMPACK_HOME;
func __op_init
/* DOCUMENT OPTIMPACK_HOME
 *   Global variable used to store the full path to the directory where OPTIMPACK
 *   software suite is installed.
 *
 * SEE ALSO:
 */
{
  /**/extern OPTIMPACK_HOME;

  /* Figure out where is "fittest.i" to define OPTIMPACK_HOME. */
  path = current_include();
  name = "OptimPack3.i";
  length = strlen(name);
  if (strpart(path, -length:0) != "/" + name) {
    error, "unexpected file name: \"" + path + "\"";
  }
  dir = strpart(path, 1:-length);
  cwd = cd(".");
  absolute_dir = cd(dir);
  cd, cwd;
  if (is_void(absolute_dir)) {
    error, "bad directory: \"" + dir + "\"";
  }
  if (strpart(absolute_dir, 0:0) != "/") {
    absolute_dir += "/";
  }
  OPTIMPACK_HOME = absolute_dir;

  /* Setup list of directories for plugins so that OPTIMPACK_HOME is searched
     first. */
  if (is_func(plug_dir)) {
    dirs = plug_dir();
    if (is_void(dirs)) {
      plug_dir, absolute_dir;
    } else {
      /* move directory in first position */
      plug_dir, grow(absolute_dir, dirs(where(dirs != absolute_dir)));
    }    
  }  
}
__op_init;
__op_init = []; /* no longer needed */

if (is_func(plug_in)) plug_in, "OptimPack3";

/*---------------------------------------------------------------------------*/

local OP_BLAS_ROW_MAJOR, OP_BLAS_COL_MAJOR;
local OP_BLAS_NO_TRANS, OP_BLAS_TRANS, OP_BLAS_CONJ_TRANS;
local OP_BLAS_UPPER, OP_BLAS_LOWER;
local OP_BLAS_NON_UNIT, OP_BLAS_UNIT;
local OP_BLAS_LEFT, OP_BLAS_RIGHT;
/* DOCUMENT OP_BLAS_ROW_MAJOR
       -or- OP_BLAS_COL_MAJOR
     Storage order for packed matrices.
   SEE ALSO:
 */
OP_BLAS_ROW_MAJOR  = 101n;
OP_BLAS_COL_MAJOR  = 102n;
OP_BLAS_NO_TRANS   = 111n;
OP_BLAS_TRANS      = 112n;
OP_BLAS_CONJ_TRANS = 113n;
OP_BLAS_UPPER      = 121n;
OP_BLAS_LOWER      = 122n;
OP_BLAS_NON_UNIT   = 131n;
OP_BLAS_UNIT       = 132n;
OP_BLAS_LEFT       = 141n;
OP_BLAS_RIGHT      = 142n;

/*---[ Level 1 BLAS-like routines ]------------------------------------------*/

extern _op_samax;
/* PROTOTYPE float op_samax(long n, float array x, long incx); */
extern _op_damax;
/* PROTOTYPE double op_damax(long n, double array x, long incx); */
extern _op_samax1;
/* PROTOTYPE float op_samax1(long n, float array x); */
extern _op_damax1;
/* PROTOTYPE double op_damax1(long n, double array x); */

extern _op_sasum;
/* PROTOTYPE float op_sasum(long n, float array x, long incx); */
extern _op_dasum;
/* PROTOTYPE double op_dasum(long n, double array x, long incx); */
extern _op_sasum1;
/* PROTOTYPE float op_sasum1(long n, float array x); */
extern _op_dasum1;
/* PROTOTYPE double op_dasum1(long n, double array x); */

extern _op_saxpy;
/* PROTOTYPE void op_saxpy(long n, float a, float array x, long incx, float
   array y, long incy); */
extern _op_daxpy;
/* PROTOTYPE void op_daxpy(long n, double a, double array x, long incx,
   double array y, long incy); */
extern _op_saxpy1;
/* PROTOTYPE void op_saxpy1(long n, float a, float array x, float array
   y); */
extern _op_daxpy1;
/* PROTOTYPE void op_daxpy1(long n, double a, double array x, double array
   y); */
extern _op_saxpy2;
/* PROTOTYPE void op_saxpy2(long n, float a, int array f, float array x,
   float array y); */
extern _op_daxpy2;
/* PROTOTYPE void op_daxpy2(long n, double a, int array f, double array x,
   double array y); */
extern _op_saxpy3;
/* PROTOTYPE void op_saxpy3(long n, float a, float array w, float array x,
   float array y); */
extern _op_daxpy3;
/* PROTOTYPE void op_daxpy3(long n, double a, double array w, double array
   x, double array y); */

extern _op_scopy;
/* PROTOTYPE void op_scopy(long n, float array x, long incx, float array y,
   long incy); */
extern _op_dcopy;
/* PROTOTYPE void op_dcopy(long n, double array x, long incx, double array
   y, long incy); */
extern _op_scopy1;
/* PROTOTYPE void op_scopy1(long n, float array x, float array y); */
extern _op_dcopy1;
/* PROTOTYPE void op_dcopy1(long n, double array x, double array y); */

extern _op_sdot;
/* PROTOTYPE float op_sdot(long n, float array x, long incx, float array y,
   long incy); */
extern _op_ddot;
/* PROTOTYPE double op_ddot(long n, double array x, long incx, double array
   y, long incy); */
extern _op_sdot1;
/* PROTOTYPE float op_sdot1(long n, float array x, float array y); */
extern _op_ddot1;
/* PROTOTYPE double op_ddot1(long n, double array x, double array y); */
extern _op_sdot2;
/* PROTOTYPE float op_sdot2(long n, int array f, float array x, float array
   y); */
extern _op_ddot2;
/* PROTOTYPE double op_ddot2(long n, int array f, double array x, double
   array y); */
extern _op_sdot3;
/* PROTOTYPE float op_sdot3(long n, float array w, float array x, float
   array y); */
extern _op_ddot3;
/* PROTOTYPE double op_ddot3(long n, double array w, double array x, double
   array y); */

extern _op_snrm2;
/* PROTOTYPE float op_snrm2(long n, float array x, long incx); */
extern _op_dnrm2;
/* PROTOTYPE double op_dnrm2(long n, double array x, long incx); */
extern _op_snrm2_1;
/* PROTOTYPE float op_snrm2_1(long n, float array x); */
extern _op_dnrm2_1;
/* PROTOTYPE double op_dnrm2_1(long n, double array x); */
extern _op_snrm2_2;
/* PROTOTYPE float op_snrm2_2(long n, int array f, float array x); */
extern _op_dnrm2_2;
/* PROTOTYPE double op_dnrm2_2(long n, int array f, double array x); */
extern _op_snrm2_3;
/* PROTOTYPE float op_snrm2_3(long n, float array w, float array x); */
extern _op_dnrm2_3;
/* PROTOTYPE double op_dnrm2_3(long n, double array w, double array x); */

extern _op_sscal;
/* PROTOTYPE void op_sscal(long n, float alpha, float array x, long incx); */
extern _op_dscal;
/* PROTOTYPE void op_dscal(long n, double alpha, double array x, long incx); */
extern _op_sscal1;
/* PROTOTYPE void op_sscal1(long n, float alpha, float array x); */
extern _op_dscal1;
/* PROTOTYPE void op_dscal1(long n, double alpha, double array x); */

extern _op_ssum;
/* PROTOTYPE float op_ssum(long n, float array x, long incx); */
extern _op_dsum;
/* PROTOTYPE double op_dsum(long n, double array x, long incx); */
extern _op_ssum1;
/* PROTOTYPE float op_ssum1(long n, float array x); */
extern _op_dsum1;
/* PROTOTYPE double op_dsum1(long n, double array x); */

extern _op_sswap;
/* PROTOTYPE void op_sswap(long n, float array x, long incx, float array y,
   long incy); */
extern _op_dswap;
/* PROTOTYPE void op_dswap(long n, double array x, long incx, double array
   y, long incy); */
extern _op_sswap1;
/* PROTOTYPE void op_sswap1(long n, float array x, float array y); */
extern _op_dswap1;
/* PROTOTYPE void op_dswap1(long n, double array x, double array y); */

extern _op_szero;
/* PROTOTYPE void op_szero(long n, float array x, long incx); */
extern _op_dzero;
/* PROTOTYPE void op_dzero(long n, double array x, long incx); */
extern _op_szero1;
/* PROTOTYPE void op_szero1(long n, float array x); */
extern _op_dzero1;
/* PROTOTYPE void op_dzero1(long n, double array x); */

extern _op_isamax;
/* PROTOTYPE long op_isamax(long n, float array x, long incx); */
extern _op_idamax;
/* PROTOTYPE long op_idamax(long n, double array x, long incx); */

/*---[ Level 2 BLAS-like routines ]------------------------------------------*/

extern _op_sgemv;
/* PROTOTYPE int op_sgemv(int trans, long m, long n, float alpha,
   float array a, long lda, float array x, long incx, float beta, float
   array y, long incy); */
extern _op_dgemv;
/* PROTOTYPE int op_dgemv(int trans, long m, long n, double alpha,
   double array a, long lda, double array x, long incx, double beta, double
   array y, long incy); */
extern _op_strmv;
/* PROTOTYPE int op_strmv(int uplo, int trans, int diag, long n, float
   array a, long lda, float array x, long incx); */
extern _op_dtrmv;
/* PROTOTYPE int op_dtrmv(int uplo, int trans, int diag, long n,
   double array a, long lda, double array x, long incx); */
extern _op_strsv;
/* PROTOTYPE int op_strsv(int uplo, int trans, int diag, long n, float
   array a, long lda, float array x, long incx); */
extern _op_dtrsv;
/* PROTOTYPE int op_dtrsv(int uplo, int trans, int diag, long n,
   double array a, long lda, double array x, long incx); */

/*---[ Level 3 BLAS-like routines ]------------------------------------------*/

extern _op_sgemm;
/* PROTOTYPE int op_sgemm(int transa, int transb, long m, long n, long
   k, float alpha, float array a, long lda, float array b, long ldb, float
   beta, float array c, long ldc); */
extern _op_dgemm;
/* PROTOTYPE int op_dgemm(int transa, int transb, long m, long n, long
   k, double alpha, double array a, long lda, double array b, long ldb,
   double beta, double array c, long ldc); */
extern _op_ssyrk;
/* PROTOTYPE int op_ssyrk(int uplo, int trans, long n, long k, float
   alpha, float array a, long lda, float beta, float array c, long ldc); */
extern _op_dsyrk;
/* PROTOTYPE int op_dsyrk(int uplo, int trans, long n, long k, double
   alpha, double array a, long lda, double beta, double array c, long ldc); */

/*---[ LAPACK-like routines ]------------------------------------------------*/

extern _op_spotf2;
/* PROTOTYPE long op_spotf2(int uplo, long n, float array a, long
   lda); */
extern _op_dpotf2;
/* PROTOTYPE long op_dpotf2(int uplo, long n, double array a, long
   lda); */

/*---[ MINPACK2-like routines ]----------------------------------------------*/

extern _op_sestsv;
/* PROTOTYPE float op_sestsv(long n, float array r,long ldr, float array
   z); */
extern _op_destsv;
/* PROTOTYPE double op_destsv(long n, double array r,long ldr, double array
   z); */
extern _op_sgqt;
/* PROTOTYPE int op_sgqt(long n, float array a, long lda, float array b,
   float delta, float rtol, float atol, long itmax, float array par, float
   array f, float array x, long array iter, float array z, float array wa1,
   float array wa2); */
extern _op_dgqt;
/* PROTOTYPE int op_dgqt(long n, double array a, long lda, double array b,
   double delta, double rtol, double atol, long itmax, double array par,
   double array f, double array x, long array iter, double array z, double
   array wa1, double array wa2); */

/*---[ Conjugate Gradient ]--------------------------------------------------*/

local OP_CG_START, OP_CG_NEWX, OP_CG_PRECOND, OP_CG_AP, OP_CG_ERROR;
/* DOCUMENT OP_CG_START   - start with no initial parameters
 *          OP_CG_RESTART - start/restart with initial parameters 
 *          OP_CG_NEWX    - new step taken
 *          OP_CG_PRECOND - caller has to to compute z = Q.r
 *          OP_CG_AP      - caller has to to compute q = A.p
 *          OP_CG_ERROR   - error   
 * SEE ALSO: op_cg
 */
OP_CG_START   =  0n;
OP_CG_RESTART =  1n;
OP_CG_NEWX    =  2n;
OP_CG_PRECOND =  3n;
OP_CG_AP      =  4n;
OP_CG_ERROR   = -1n;

// FIXME: |r|^2 is stored in RHO
// FIXME: criterion sometimes increase

func op_cg(a, b, x, precond=, maxiter=, maxeval=, verb=, view=, atol=, rtol=)
{
  type = structof(b);
  if (type == double) {
    cg = (precond ? _op_dpcg : _op_dlcg);
  } else {
    if (type == char || type == short || type == int || type == long) {
      b = float(b);
      type = float;
    }
    if (type == float) {
      cg = (precond ? _op_spcg : _op_slcg);
    } else {
      error, "expecting non-complex numerical array for RHS vector B";
    }
  }
  dims = dimsof(b);
  if (is_void(x)) {
    state = OP_CG_START;
    x = array(type, dims);
  } else {
    xdims = dimsof(x);
    if (numberof(xdims) != numberof(dims) || anyof(xdims != dims)) {
      error, "bad dimension list for X";      
    }
    xtype = structof(x);
    if (xtype != type) {
      if (xtype == double) {
        write, format="%s\n", "WARNING - converting X to single precision";
      } else if (xtype != char && type != short && type != int &&
                 type != long && xtype != float) {
        error, "bad data type for X";
      }
      x = type(x);
    }
    state = (anyof(x) ? OP_CG_RESTART : OP_CG_START);
  }
  p = q = r = b; // FIXME: slight optimization possible here (eq_nocopy)
  if (precond) {
    z = p;
  }
  rho = array(type, 4);
  n = numberof(b);
  if (is_void(maxiter)) {
    maxiter = n;
  }
  iter = -1;
  eval = 0;
  if (verb) {
    write, format="%s\n%s\n",
      " Iter    Residual Norm   CPU [s]      Alpha       Beta"
      "--------------------------------------------------------";
    format = "%5d %16.8e %9.3f  %10.2e  %10.2e\n";
    elapsed = array(double, 3);
    timer, elapsed;
    cpu_start = elapsed(1);
  }
  for (;;) {
    /* Call reverse commnunication version of conjugate gradient
       routine, also make sure STATE (which is passed by address) is
       of proper type. */
    state = int(state)
    if (precond) {
      cg, n, p, q, r, x, z, rho, state;
    } else {
      cg, n, p, q, r, x, rho, state;
    }

    /* Behave according to output value of STATE. */
    if (state == OP_CG_AP) {
      q = type(a(p)); // FIXME: type(linop_apply(a, p, 0))
      ++eval;
    } else if (state == OP_CG_PRECOND) {
      z = type(precond(r)); // FIXME: type(linop_apply(precond, r, 0))
    } else if (state == OP_CG_NEWX || state == OP_CG_FINISH) {
      ++iter;
      rnrm2 = op_nrm2(r);
      if (iter == 0) {
        if (! is_void(rtol)) {
          if (is_void(atol)) {
            atol = rtol*rnrm2;
          } else {
            atol = min(atol, rtol*rnrm2);
          }
        } else if (is_void(atol)) {
          atol = 1e-6*rnrm2;
        }
      } else if (iter % n == 0) {
        r = b;
        state = OP_CG_RESTART;
      }
      convergence = (rnrm2 <= atol);
      finish = (convergence || iter >= maxiter);
      prnt = (verb && (finish || ! (iter % verb)));
      if (prnt) {
        timer, elapsed;
        cpu = elapsed(1) - cpu_start;
        write, format=format, iter, rnrm2, cpu, rho(3), rho(4);
        if (view) {
          view, x;
        }
      }
      if (finish) {
        if (! convergence) {
          write, format="%s\n", "WARNING - too many iterations";
        }
        return x;
      }
    } else {
      error, "error in PCG";
    }
  }
}

extern _op_slcg;
/* PROTOTYPE void op_slcg(long n, float array p, float array q,
                          float array r, float array x,
                          float array rho, int array state); */
extern _op_dlcg;
/* PROTOTYPE void op_dlcg(long n, double array p, double array q,
                          double array r, double array x,
                          double array rho, int array state); */
extern _op_spcg;
/* PROTOTYPE void op_spcg(long n, float array p, float array q,
                          float array r, float array x, float array z,
                          float array rho, int array state); */
extern _op_dpcg;
/* PROTOTYPE void op_dpcg(long n, double array p, double array q,
                          double array r, double array x, double array z,
                          double array rho, int array state); */

LONG_MAX = long(2.0^(8.0*sizeof(long) - 1.0) - 1.0);
LONG_MIN = -1L - LONG_MAX;

func op_same_dimlist(adims, bdims)
/* DOCUMENT op_same_dimlist(adims, bdims);
 *   Retuns true if ADIMS and BDIMS are the same dimension lists; false
 *   otherwise.  This function does not check that ADIMS and BDIMS are
 *   valid dimension lists.
 *    
 * SEE ALSO: dimsof.
 */
{
  return (numberof(adims) == numberof(bdims) && allof(adims == bdims));
}

// FIXME: speedup by computing sparse A'.W.A matrix
func op_cbspline_fit(x, y, step=, offset=, weight=, sart=)
{
  m = numberof(x);
  dims = [1, m];
  if (! op_same_dimlist(dimsof(x), dims) ||
      ! op_same_dimlist(dimsof(y), dims)) {
    error, "A and Y must be vectors of same length";
  }

  /* Convert X in coordinates w.r.t. spline grid. */
  if (is_void(step)) {
    step = 1.0;
  } else if (step <= 0.0) {
    error, "bad value for STEP";
  }
  if (is_void(offset)) {
    offset = avg(x);
  }
  if (offset != 0.0) {
    x -= offset;
  }
  if (step != 1.0) {
    x *= (1.0/step);
  }

  /*
   *                  g(k)  x
   *  -----+------------+---*--------+----> coordinates
   *       |<-- step -->|
   *
   *  where:
   *
   *      g(k) = offset + k*step
   *
   *  is the lower bound of X on the grid.  Index K is given by:
   *
   *      k = floor((x - offset)/step)
   *        = floor(t)
   *
   *  where:
   *
   *      t = (x - offset)/step
   *
   *  are the coordinates w.r.t. to the grid units.  If H is the half-width
   *  of the spline, then non-zero spline coefficients are for:
   *
   *      k - h + 1, k - h + 2, ..., k + h
   *
   *  if X exactly match g(k), then non-zero spline coefficients are for:
   *
   *      k - h + 1, k - h + 2, ..., k + h - 1
   *
   *  We choose to not account for all splines which would have non-zero
   *  coefficients but only for the nodes such that:
   *
   *      floor(t) <= k <= ceil(t)
   *
   *  which is similar to choosing particular derivative conditions at the
   *  end points of the interval.
   */

  /* Compute number of splines. */
  kmin = floor(min(x));
  kmax = ceil(max(x));
  n = kmax - kmin + 1.0; /* number of spline nodes */
  if (n > m) {
    error, "too many degrees of freedom -- the grid spacing (STEP) is too small";
  }
  n = long(n); /* now we can safely convert N into an integer */
  
  /* Build the sparse matrix. */
  k = floor(x); /* spline node index at lower bound */
  u = x - k; /* offset w.r.t. lower bound */
  k = long(k + (1.0 - kmin)); /* convert K into valid Yorick index */
  write, n;
  stat, k;
  coefs = [];
  row_indices = [];
  col_indices = [];
  h = 2; /* half-width of a cubic spline basis function */
  for (j = 1 - h; j <= h; ++j) {
    s = op_cbspline(u - j);
    i = where((s > 0.0)&(k >= 1 - j)&(k <= n - j));
    if (is_array(i)) {
      grow, coefs, s(i);
      grow, row_indices, i;
      grow, col_indices, k(i) + j;
    }
  }
  s = sparse_matrix(coefs, m, row_indices, n, col_indices);
  //return s;

  local _op_cbspline_fit_p;
  local _op_cbspline_fit_x, _op_cbspline_fit_y;
  local _op_cbspline_fit_s, _op_cbspline_fit_weight;
  eq_nocopy, _op_cbspline_fit_x, x;
  eq_nocopy, _op_cbspline_fit_y, y;  
  eq_nocopy, _op_cbspline_fit_s, s;
  
  if (sart) {
     /* Compute diagonal SART preconditioner. */
    if (is_void(weight)) {
      u = s(s(array(1.0, n)), 1);
    } else {
      u = s(weight*s(array(1.0, n)), 1);
    }
    _op_cbspline_fit_p = 1.0/u;
    precond=_op_cbspline_fit_precond;
  }
  stat,y;
  if (is_void(weight)) {
    worker = _op_cbspline_fit_worker1;
    b = s(y, 1);
  } else {
    eq_nocopy, _op_cbspline_fit_weight, weight;
    worker = _op_cbspline_fit_worker2;
    b = s(weight*y, 1);
  }
  stat,b;
  w = op_cg(worker, b, precond=precond,
            maxiter=2*n, maxeval=, verb=1, view=_op_cbspline_fit_view,
            atol=0, rtol=0);
  return w;
  
}
func _op_cbspline_fit_view(x)
{
  window, 51;
  fma;
  plp, _op_cbspline_fit_y, _op_cbspline_fit_x, color="blue";
  plg, _op_cbspline_fit_s(x), _op_cbspline_fit_x, color="red";
  pause, 1;
}
func _op_cbspline_fit_worker1(x)
{
  return _op_cbspline_fit_s(_op_cbspline_fit_s(x), 1);
}
func _op_cbspline_fit_worker2(x)
{
  return _op_cbspline_fit_s(_op_cbspline_fit_weight*_op_cbspline_fit_s(x), 1);
}
func _op_cbspline_fit_precond(r)
{
  return _op_cbspline_fit_p*r;
}
# if 0
include,"./OptimPack3.i";
x = span(-100, 300, 500);
y0 = sin(0.1*x)*exp(0.01*x);
y = y0 + 3*random_n(dimsof(x));
plg,y0,x;
plp,y,x;
op_cbspline_fit, x, y, step=10;

stdev = sqrt(9.0 + 2.0*abs(y0));
y = y0 + stdev*random_n(dimsof(x));
weight = 1.0/(stdev*stdev);
#endif

local OP_CBSPLINE_FWHM;
func op_cbspline(x)
/* DOCUMENT op_cbspline(x)
 *   Returns the profile of the cubic B-spline centered at 0 onto a regular
 *   grid at positions X (in units of the grid) and defined by:
 *
 *                      2/3 + (X/2 - 1)*X^2     where abs(X) <= 1
 *     op_cbspline(X) = 1/6 * (2 - X)^3         where 1 <= abs(X) <= 2
 *                      0                       where abs(X) >= 2
 *
 *   This cubic B-spline has the following properties:
 *   (1) op_cbspline(x) is exactly zero outside (including boundaries) the
 *       support [-2, +2].  op_cbspline(x) and its first and second derivatives
 *       are continuous functions everywhere and are piecewise polynomials.
 *   (2) op_cbspline is an even unimodal bell shaped function:
 *         op_cbspline(x) = op_cbspline(-x)
 *         max(op_cbspline(x)) = op_cbspline(0) = 2/3
 *   (3) The full width at half maximum of this cubic B-spline is:
 *         OP_CBSPLINE_FWHM = 4/3 * (1 + 2*cos((pi + atan(3*sqrt(7)))/3))
 *                          = 1.444703448928752590331004298512891328576...
 *       this value is stored in global variable OP_CBSPLINE_FWHM.
 *   (4) The cubic B-spline is normalized:
 *
 *           _ +Inf                  _ +2
 *          |                       |
 *          |  op_cbspline(x) dx =  |  op_cbspline(x) dx = 1
 *         _|                      _|
 *           -Inf                   -2
 *           
 *       and: op_cbspline(-1) + op_cbspline(0) + op_cbspline(+1) = 1
 *
 *
 *  SEE ALSO: op_cbspline_prime, op_cbspline_second, op_cbspline_integral.
 */
{
  x = abs(x);
  y = array(double, dimsof(x));
  if ((is_array((j = where(x < 2.0))))) {
    test = ((x = x(j)) <= 1.0);
    if (is_array((i = where(test)))) {
      u = x(i);
      y(j(i)) = (0.5*u - 1.0)*u*u + (2.0/3.0);
    }
    if (is_array((i = where(! test)))) {
      u = 2.0 - x(i);
      y(j(i)) = (1.0/6.0)*u*u*u;
    }
  }
  return y;
}
OP_CBSPLINE_FWHM = 1.444703448928752590331004298512891328576;

/*---------------------------------------------------------------------------*/

func op_gqtpar(a, b, delta, &par, &f, &iter, &status, rtol=, atol=, itmax=,
               verb=)
/* DOCUMENT x = func op_gqtpar(a, b, delta, par, f, iter, status)
 *
 *   Given an N by N symmetric matrix A, an N-vector B, and a positive
 *   number DELTA, this function determines a vector X which approximately
 *   minimizes the quadratic function:
 *
 *       f(x) = (1/2)*x'.A.x + b'.x
 *
 *   subject to the Euclidean norm constraint:
 *
 *       norm(x) <= delta.
 *
 *   This function computes an approximation X and a Lagrange multiplier
 *   PAR such that either PAR is zero and
 *
 *        norm(x) <= (1 + rtol)*delta,
 *
 *   or PAR is positive and
 *
 *        abs(norm(x) - delta) <= rtol*delta.
 *
 *   If XSOL is the exact solution to the problem, the approximation X
 *   satisfies:
 *
 *        f(x) <= ((1 - rtol)^2)*f(xsol)
 *
 *
 * ARGUMENTS
 *
 *   A is a real array of dimension N by N.  The full upper triangle of A
 *     must contain the full upper triangle of the symmetric matrix A.
 *
 *   B is a real array of dimension N which specifies the linear term
 *     in the quadratic approximation.
 *
 *   DELTA is a positive real value equal to the bound on the Euclidean
 *     norm of X.
 *
 *   PAR is a real scalar.  On entry, PAR is an initial estimate of the
 *     Lagrange multiplier for the constraint norm(x) <= delta.  If PAR
 *     is undefined on entry, PAR=0.0 is tried first.  On exit, PAR
 *     contains the final estimate of the multiplier.
 *
 *   F is a double precision variable.  On entry, F need not be specified.
 *     On exit, F is set to f(X) at the output X.
 *
 *   STATUS is an integer variable.  On entry STATUS need not be specified.
 *     On exit STATUS is set as follows:
 *
 *        status = 1  The function value f(x) has the relative
 *                  accuracy specified by rtol.
 *
 *        status = 2  The function value f(x) has the absolute
 *                  accuracy specified by atol.
 *
 *        status = 3  Rounding errors prevent further progress.
 *                  On exit x is the best available approximation.
 *
 *        status = 4  Failure to converge after itmax iterations.
 *                  On exit x is the best available approximation.
 *
 * KEYWORDS
 *
 *   RTOL is the relative accuracy desired in the solution. Convergence
 *     occurs if:
 *
 *          f(x) <= ((1 - rtol)^2)*f(xsol)
 *
 *     By default RTOL=0.05; otherwise the value for RTOL must be a
 *     positive real scalar.
 *
 *   ATOL is the absolute accuracy desired in the solution. Convergence
 *     occurs when:
 *
 *          norm(x) <= (1 + rtol)*delta
 *
 *          max(-f(x), -f(xsol)) <= atol
 *
 *     By default ATOL=0.01; otherwise the value for ATOL is a real scalar.
 *
 *   ITMAX is the maximum number of iterations.  By default ITMAX=10;
 *     otherwise the value for ITMAX is an integer scalar.
 *
 *
 * RESULT
 *
 *   X is a double precision array of dimension n.
 *     On entry x need not be specified.
 *     On exit x is set to the final estimate of the solution.
 *
 *
 * REFERENCES
 *
 *   [1] Mor, J.J. & Sorensen, D.C., "Computing A Trust Region Step",
 *       SIAM J. Sci. Stat. Comp., vol. 4, pp. 553-572, 1983.
 */
{
  if (! is_array(a) || (dims = dimsof(a))(1) != 2 || dims(2) != dims(3)) {
    error, "matrix A must be a N-by-N array";
  }
  n = dims(2);
  if ((s = structof(a)) == double) {
    real = double;
  } else if (s == float || s == long || s == int || s == short || s == char) {
    real = float;
  } else {
    error, "bad data type for matrix A";
  }

  if (! is_array(b) || (dims = dimsof(b))(1) != 1 || dims(2) != n) {
    error, "vector B must be a N array";
  }
  if ((s = structof(b)) == double) {
    real = double;
    op = _op_dgqt;
  } else if (s == float || s == long || s == int || s == short || s == char) {
    op = (real == double ? _op_dgqt : _op_sgqt);
  } else {
    error, "bad data type for vector B";
  }
  a = real(a); /* force private copy and convert to correct data type */
  b = real(b); /* force private copy and convert to correct data type */
  f = real(0);
  iter = long(0);
  x = array(real, n);
  z = array(real, n);
  wa1 = array(real, n);
  wa2 = array(real, n);
  if (is_void(par)) par = real(0); else par = real(par);
  if (is_void(itmax)) itmax = 10;
  if (is_void(rtol)) rtol = 0.05;
  if (is_void(atol)) atol = 0.01; /* FIXME: */
  
  status = op(n, a, n, b, delta, rtol, atol, itmax, par, f, x, iter, z, wa1, wa2);
  if (verb) write, format="OP_GQTPAR: %s\n", op_gqtpar_info(status);
  return x;
}
func op_gqtpar_info(status)
{
  if (status == 1) {
    return "The function value f(x) has the relative accuracy specified by rtol.";
  } else if (status == 2) {
    return "The function value f(x) has the absolute accuracy specified by atol.";
  } else if (status == 3) {
    return "Rounding errors prevent further progress.  On exit x is the best available approximation.";
  } else if (status == 4) {
    return "Failure to converge after itmax iterations.  On exit x is the best available approximation.";
  } else {
    return "Unknown status value. (BUG?)";
  }
}

/*---------------------------------------------------------------------------*/
func op_nllsq(f, x, extra=, weight=, scale=,
              fit=, eps=, tiny=,
              prec=, verb=, maxeval=, maxstep=,
              delta=, eta=, rho=, gamma=)
/* DOCUMENT op_nllsq(f, x)
 *   Non-linear least squares.
 *   
 *   F = user defined function, return a vector of residuals given the
 *       parameters.
 *   
 *   X = initial set of parameters.
 *   
 *   EXTRA = any extra arguments required by F.  The calling sequence is
 *       F(X, EXTRA) if EXTRA is non-nil or just F(X) if EXTRA is nil.
 *     
 *   SCALE = scaling of parameters.  Specified as the name of the strategy
 *       to use or as an array of scaling factor(s):
 *
 *           "none" or 1.0    - no scaling
 *           "default" or nil - scaling by the magnitude of the parameters
 *           "max"            - variant of "default"
 *           "abs"            - variant of "default"
 *
 *       The parameters are scaled by dividing them by their SCALE;
 *       this is intended to have all rescaled parameters have
 *       approximatively the same magnitude to limit the
 *       ill-conditioning of the criterion to minimize.  If SCALE is
 *       nil, automatic scaling will be used (the "default" strategy).
 *       If SCALE is given by numerical value, it must be an array
 *       conformable with the array parameters X.
 *     
 *   DELTA = initial bound for the scaled step.
 *   
 *   WEIGHT = diagonal weights the function to be minimized
 *      is sum(F(X)^2) or sum(WEIGHT*F(X)^2)
 *    
 *   PREC = required relative precision for the parameters
 *   
 *   FIT = indices of parameters to fit, default is all.  For instance, if you
 *      want to fit the second and last parameters, you can use: FIT=[1,0].
 *   
 *   TINY = small value, all parameters with their absolute value
 *      below this value will scale the same.  Must be strictly greater
 *      than zero.
 */
{
  nevals = 0; /* number of function evaluations */
  nsteps = 0; /* number of sucessful steps */
  nrejects = 0; /* number of unsuccessful steps */

  /* parse trust region settings */
  if (is_void(delta)) {
    delta = 0.1;
  } else if (delta <= 0.0) {
    error, "bad value for initial trust region radius DELTA";
  }
  if (is_void(eta)) {
    eta = 0.1;
  } else if (eta <= 0.0 || eta >= 1.0) {
    error, "bad value for keyword ETA";
  }
  if (is_void(rho)) {
    rho1 = 0.25;
    rho2 = 0.75;
    rho3 = 1.25;
  } else if (numberof(rho) != 3 || (rho1 = rho(1)) <= 0.0 ||
             (rho2 = rho(2)) >= 1.0 || (rho3 = rho(3)) <= 1.0 ||
             rho1 >=  rho2) {
    error, ("keyword RHO must have 3 values such that: "
            + "0 < RHO(1) < RHO(2) < 1 < RHO(3)");
  }
  if (is_void(gamma)) {
    gamma1 = 0.5;
    gamma2 = 2.0;
    gamma3 = 1.0;
  } else if (numberof(gamma) != 3 || (gamma1 = gamma(1)) <= 0.0 ||
             gamma1 >= 1.0 || (gamma2 = gamma(2)) <= 1.0 ||
             (gamma3 = gamma(3)) <= 0.0) {
    error, ("keyword GAMMA must have 3 values such that: "
            + "0 < GAMMA(1) < 1 < GAMMA(2) and GAMMA(3) > 0");
  }

  
  if (is_void(eps)) eps = 1e-5;
  else if (eps <= 0.0 || eps >= 1.0) error, "bad value for keyword EPS";
  if (is_void(tiny)) tiny = 1e-12;
  else if (tiny <= 0.0) error, "bad value for keyword TINY";
  if (eps*tiny <= 0.0) error, "TINY and/or EPS are too small";
  
  /* user defined function takes extra parameters? */
  use_extra = ! is_void(extra);

  /* caller has set limits for number of iterations/evaluations? */
  use_maxeval = ! is_void(maxeval);
  use_maxstep = ! is_void(maxstep);
  
  /* number of parameters, initial residuals and number of residuals */
  if (structof(x) == complex) error, "unsupported complex parameters";
  eq_nocopy, x0, double(x);
  nx = numberof(x);
  xdims = dimsof(x);
  f0 = (use_extra ? f(x0, extra) : f(x0))(*);
  ++nevals;
  m = numberof(f0);
  err0 = sum(f0*f0);

  /* list of parameters to fit */
  if (is_void(fit)) {
    fit = indgen(nx);
  } else {
    /* check/fix FIT argument */
    if (! is_integer(fit)) {
      error, "expecting integer value for keyword FIT";
    }
    if (min(fit) <= 0) {
      fit += nx*(fit <= 0);
    }
    if (min(fit) <= 0 || max(fit) > nx) {
      error, "out of range indices in FIT keyword";
    }
    test = array(int, nx);
    test(fit) = 1n;
    fit = where(test);
  }
  n = numberof(fit);
  if (n <= 0) error, "no parameters to fit";
  // FIXME: account for the weights
  if (n > m) error, "too many parameters to fit";

  
  /* scaling of parameters */
  if (is_void(scale)) {
    auto_scale = 1;
  } else if (is_string(scale)) {
    auto_scale = -1;
    if (is_scalar(scale)) {
      if (scale == "none") {
        auto_scale = 0;
      } else if (scale == "default") {
        auto_scale = 1;
      } else if (scale == "max") {
        auto_scale = 2;
      } else if (scale == "abs") {
        auto_scale = 3;
      }
    }
    if (auto_scale < 0) {
      error, "bad balue for keyword SCALE";
    }
  } else {
    if (min(scale) <= 0.0) {
      error, "bad SCALE";
    }
    scale = (scale*array(1.0, xdims))(fit);
    auto_scale = 0;
  }

  if (is_void(prec)) prec = 1e-8;
  smin = prec; // FIXME:

  /* Trust-region settings. */
  alpha1 = 0.5;
  alpha2 = 2.0;
  alpha3 = 1.01;
  eta1 = 0.01;
  eta2 = 0.95;
  eta3 = 1.05;
  rtol = 0.05;
  atol = 1e-5; // FIXME:
  itmax = 10;


  status = 0;
  beta = 1.0/eps;
  actual_reduct = 0.0;
  predicted_reduct = 0.0;
  par = 0.0;
  for (;;) {

    if (verb) {
      if (nsteps < 1) {
        hline = "-------------------------------------------------------------------------------";
        write, format="%s\n%s\n%s\n",
          hline,
          "   STEP    EVAL  REJECT           PENALTY    ACTUAL   PREDICT    DELTA   LAMBDA",
          hline;
      }
      write, format="%7d %7d %7d%18.10e %9.1e %9.1e%9.1e%9.1e\n",
        nsteps, nevals, nrejects, err0, actual_reduct, predicted_reduct, delta, par;
    }
    if (status) {
      break;
    }
    if (use_maxstep && nsteps >= maxstep) {
      status = 2; /* too many steps */
      break;
    }

    /* update parameter scaling */
    if (auto_scale) {
      if (auto_scale == 1) {
        scale = abs(x0(fit));
        scale += ! scale;
      } else if (auto_scale == 2) {
        scale = max(tiny, abs(x0(fit)));
      } else {
        scale = abs(tiny, x0(fit));
      }
    }
    
    /* compute partial derivatives of the residuals w.r.t. the scaled
       parameters by forward differences */
    pder = array(double, m, n);
    for (j=1 ; j<=n ; ++j) {
      x1 = x0;
      x1(fit(j)) += eps*scale(j); /* forward difference */
      f1 = (use_extra ? f(x1, extra) : f(x1))(*);
      ++nevals;
      if (use_maxeval && nevals >= maxevals) {
        status = 1; /* too many function evaluation */
        break;
      }
      pder(,j) = beta*(f1 - f0);
    }
    if (status) {
      break;
    }
  
    /* compute local quadratic approximation (times 1/2) */
    b = pder(+,)*f0(+);
    a = pder(+,)*pder(+,);
      
    /* loop until DELTA sufficiently small to have a good local
       approximation of the function */ 
    for (;;) {
      
      /* compute a bounded scaled step */
      local iter0, status0, dq;
      s = op_gqtpar(a, b, delta, par, dq, iter0, status0,
                    rtol=rtol, atol=atol, itmax=itmax, verb=0);

      /* compute new solution, and corresponding penalty */
      x1 = x0;
      x1(fit) += scale*s;
      f1 = (use_extra ? f(x1, extra) : f(x1))(*);
      ++nevals;
      err1 = sum(f1*f1);
      predicted_reduct = (2.0*dq);
      actual_reduct = (err1 - err0);
      snorm = _op_dnrm2(n, s, 1);
      if (snorm <= smin) {
        status = 4; /* convergence */
        break;
      } else if (predicted_reduct >= 0.0) {
        status = 3; /* no more improvement possible */
        break;
      } else if (use_maxeval && nevals >= maxevals) {
        status = 1; /* too many function evaluation */
        break;
      }

      /* update trust region size (see op_trust_region) */
      rho = actual_reduct/predicted_reduct;
      if (rho < rho1) {
        delta = gamma1*min(snorm, delta);
      } else if (par != 0.0) {
        /* trust region bound is active */
        if (rho2 <= rho && rho <= rho3) {
          delta *= gamma2;
        } else if (rho > rho3) {
          delta *= gamma3;
        }
      }
      if (rho < eta) {
        /* reject the step */
        ++nrejects;
      } else {
        /* accept the step */
        break;
      }
    }

    if (actual_reduct < 0.0) {
      /* take the step */
      eq_nocopy, x0, x1;
      eq_nocopy, f0, f1;
      err0 = err1;
      ++nsteps;
    } else if (status) {
      /* avoid printing information if termination and no improvement
         made */
      break;
    }
  }
  if (verb) {
    write, format="%s\n", hline;
  }
  
  write, format="termination with status = %d\n", status;
  return x0;
}

local op_trust_region;
/* DOCUMENT Trust region strategy
 *
 *   In non-linear optimization, the trust region method consists in
 *   restricting the size of the Newton step (derived from a local
 *   quadratic approximation of the penalty) to be not longer than the
 *   radius of the so-called trust region.  This radius must be tuned
 *   so that it is as large as possible (to allow for large steps and
 *   therefore achieve fast convergence) whereas it is small enough
 *   for the quadratic approximation to be a good model of the true
 *   penalty.
 *
 *   The updating of the trust region size is done so follows:
 *
 *     1. A parameter step is computed for the current quadratic
 *        approximation and trust region radius.
 *
 *     2. The ratio RHO of the penalty reduction predicted by the
 *        quadratic approximation divided by the actual reduction of
 *        the true penalty if the step is taken is computed.  The
 *        value of RHO is then used to estimate how good is the
 *        quadratic approximation model.
 *
 *     3. If RHO is smaller than a threshold ETA, then the parameter
 *        step is rejected; otherwise the parameter step is accepted.
 *        ETA = 0.1 is the default value in the implemented algorithm.
 *
 *     4. The trust region radius DELTA is updated as follows: if RHO
 *        is too small (say RHO < RHO_1, with typically RHO_1 = 0.25)
 *        the quadratic approximation is too bad and the trust region
 *        radius must be reduced (this is done by multiplying DELTA by
 *        a factor 0 < GAMMA_1 < 1, by default GAMMA_1 = 0.5); else if
 *        the trust region constraint apply and if the actual
 *        reduction is comparable to the predicted one (say RHO_2 <=
 *        RHO <= RHO_3, with RHO_1 < RHO_2 < 1 < RHO_3), the trust
 *        region radius can be augmented (this is done by multiplying
 *        DELTA by a factor GAMMA_2 > 1, by default GAMMA_2 = 2.0), if
 *        RHO > RHO_3 the trust region radius is multiplied by
 *        GAMMA_3.  Preferably, GAMMA_3 <= 1 which gives a reduction,
 *        the reason to not augment the trust region size when ETA >
 *        ETA_3, is that the reduction is much better than expected,
 *        so the step can be accepted although the quadratic
 *        approximation is not a good model.
 *
 *   To summarize (STPLEN is the norm of the current trial step):
 *
 *       if RHO >= ETA then
 *           accept the step
 *       else
 *           reject the step
 *       end
 *       if RHO < RHO_1 then
 *           reduce the trust region radius:
 *           DELTA = min(DELTA, STPLEN)*GAMMA_1
 *       else if trust region constraint active then
 *           if RHO_2 <= RHO <= RHO_3 then
 *               augment the trust region radius:
 *               DELTA = DELTA*GAMMA_2
 *           else if RHO > RHO_3 then
 *               DELTA = DELTA*GAMMA_3
 *           end
 *       end
 *
 *   The values of ETA, RHO_1, RHO_2, RHO_3, GAMMA_1, GAMMA_2 and GAMMA_3
 *   must verify:
 *
 *       0 < ETA <= RHO_1
 *       0 < RHO_1 < RHO_2 < 1 < RHO_3
 *       0 < GAMMA_1 < 1
 *       1 < GAMMA_2
 *       0 < GAMMA_3
 *
 *   The default settings are:
 *       ETA = 0.1         threshold for accepting the step
 *       RHO_1 = 0.25      threshold for reducing the trust region size
 *       RHO_2 = 0.75      inferior bound for augmenting the trust region size
 *       RHO_3 = 1.25      superior bound for augmenting the trust region size
 *       GAMMA_1 = 0.5     factor for reducing the trust region size
 *       GAMMA_2 = 2.0     factor for augmenting the trust region size
 *       GAMMA_3 = 1.0     factor for 
 *       
 */
/*---------------------------------------------------------------------------*/
func op_tr_update(a, b)
/* DOCUMENT 
 *
 * REFERENCES
 *   [1] Walmag, J.M.B. & Delhez, .J.M., "A note on trust-region radius
 *       update", SIAM Journal on Optimization, vol. 16, pp. 548-562,
 *       2005.
 *
 * SEE ALSO
 *   op_gqtpar
 */
{
  local f, iter, status;
  x = op_gqtpar(a, b, delta, par, f, iter, status, rtol=, atol=, itmax=);


  eta = (f_new - f_old)/df_quad;

  alpha1 = 0.5;
  alpha2 = 2.0;
  alpha3 = 1.01;
  eta1 = 0.01;
  eta2 = 0.95;
  eta3 = 1.05;

  if (strategy == 1) {
    /* 0 < eta1 <= eta2 < 1 */
    /* alpha1 < 1 < alpha2 */
    if (eta < eta1) {
      delta *= alpha1;
    } else if (eta < eta2) {
      delta = delta;
    } else {
      delta *= alpha2;
    }
  } else if (strategy == 2) {
    /* 0 < eta1 <= eta2 < 1 < eta3 */
    /* alpa1 < 1 < alpha3 < alpha2 */
    if (eta < eta1) {
      delta *= alpha1;
    } else if (eta < eta2) {
      delta = delta;
    } else if (eta < eta3) {
      delta *= alpha2;
    } else {
      delta *= alpha3;
    }
  } else if (strategy == 3) {
  } else if (strategy == 4) {
    /* Self-adaptative trust-region algorithm. */
    /* original algorithm is with ETA3 = 1 */
    if (eta <= 0.0) {
      alpha = alpha1;
    } else if (eta < eta2) {
      alpha = alpha1 + (1.0 - alpha1)*(eta/eta2)^2;
    } else if (eta < eta3) {
      alpha = alpha3 + (alpha2 - alpha3)*exp(-((eta - eta3)/(eta2 - eta3))^2);
    } else {
      alpha = alpha2*(2.0 - exp(1 - eta));
    }
  }
  
}

/*---------------------------------------------------------------------------*/
/* UTILITIES */

local op_enorm;
extern op_nrm2;
/* DOCUMENT op_nrm2(x);
       -or- op_enorm(x);
     Returns Euclidean (L-2) norm of X which is sqrt(sum(X*X))
     but avoiding overflows.
   SEE ALSO: op_dot.
 */
op_enorm = op_nrm2;

extern op_dot;
/* DOCUMENT op_dot(x, y);
     Returns dot product of X and Y which is sum(X*Y) but using
     BLAS-like comiled functions.  The arguments must have the same
     dimension lists.
   SEE ALSO: op_nrm2.
 */

/*---------------------------------------------------------------------------*
 * Local Variables:                                                          *
 * mode: yorick                                                              *
 * tab-width: 8                                                              *
 * coding: latin-9                                                           *
 * End:                                                                      *
 *---------------------------------------------------------------------------*/
